/*
 * author: Darren Bounds <darren@sfhn.net>
 * copyright: Copyright (C) 2002 by Darren Bounds
 * license: This software is under GPL version 2 of license
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * packit official page at http://www.intrusense.com/~dbounds
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "../include/packit.h" 
#include "../include/inject.h"
#include "../include/error.h"

u_char *
packit_retrieve_rand_v4_addr()
{
    int oct, i; 
    u_char *ip;
 
    ip = malloc(sizeof(u_char) * 16);

    for(i = 1; i < 5; i++)
    {
        srand(time(0) ^ getpid() * i);
        oct = rand() % 255;

        if(i != 1)
            snprintf(ip, 16, "%s.%d", ip, oct);
        else
            snprintf(ip, 16, "%d", oct);
    }

    return ip;
}

int
packit_build_ethernet_addr(char *ethstr, int sd)
{
     int i;
     long hex16;
     char *sep, *eptr, *brkb;

     sep = ":";
     i = 0;

     for (eptr = strtok_r(ethstr, sep, &brkb);
          eptr;
          eptr = strtok_r(NULL, sep, &brkb))
     {
         hex16 = strtol(eptr, 0, 16);

         if(sd == 1)
             ehdr_o.eh_saddr[i] = hex16;
         else if(sd == 2)
             ehdr_o.eh_daddr[i] = hex16;
       
         i++;
     }

     return 1;
}

void
enter_packet_foundry(int init_type)
{
    int hex16;
    int hdrlen = 0;
    int ip_proto = 0;
    int rand_id = 0;
    int packet_c;
    char proto[5], flags[7];
    libnet_t *packet_d;
    
    struct libnet_ether_addr *enet; 
    struct libnet_stats ln_stats;

    proto[0] = '\0';
    flags[0] = '\0';

    fprintf(stderr, "Mode: Packet Injection\n\n");

    if(burst > 100)
    {
        fprintf(stderr, "\nError: Maximum burst rate is 100\n");
        exit(EXIT_FAILURE);
    }

    if(payload != NULL)
        payload_s = strlen(payload); 

    if((packet_d = libnet_init(init_type, device, errbuf)) == NULL)
    {
        fprintf(stderr, "\nError: Unable to initialize packet injection: %s\n", errbuf);
        exit(EXIT_FAILURE);
    }

    if(init_type == 0)
    {
        if(ehdr_o.str_saddr == NULL)
        {
            if((enet = libnet_get_hwaddr(packet_d)) == NULL)
            {
                 fprintf(stderr, "\nError: Unable to determine ethernet address: %s\n", libnet_geterror(packet_d));
                 exit(EXIT_FAILURE);
            }

            strcpy(ehdr_o.eh_saddr, enet->ether_addr_octet);

                for (hex16 = 0; hex16 < 6; hex16++)
                    ehdr_o.eh_saddr[hex16] = enet->ether_addr_octet[hex16];
        }
        else
        {
            packit_build_ethernet_addr(ehdr_o.str_saddr, 1); 
        }

        if(ehdr_o.str_daddr == NULL)
             fprintf(stderr, "Warning: Using null destination ethernet address. Packets may not reach destination.\n\n");
        else
             packit_build_ethernet_addr(ehdr_o.str_daddr, 2);
    }
     
    if(iphdr_o.version == 4)
    {
        if(iphdr_o.daddr == NULL)
        {
             fprintf(stderr, "\nError: No destination IP address defined\n");
             exit(EXIT_FAILURE);
        }

        if(iphdr_o.rand_saddr)
           iphdr_o.saddr = packit_retrieve_rand_v4_addr();

        if(iphdr_o.saddr == NULL)
        {
            if((iphdr_o.v4saddr_nbo = libnet_get_ipaddr4(packet_d)) == -1)
            {
                fprintf(stderr, "\nError: Unable to determine local IP address: %s\n", libnet_geterror(packet_d)); 
                exit(EXIT_FAILURE);
            }
            else
            {
                iphdr_o.saddr = libnet_addr2name4(iphdr_o.v4saddr_nbo, 0);     
            }
        }

        if((iphdr_o.v4saddr_nbo = libnet_name2addr4(packet_d, iphdr_o.saddr, 1)) == -1)
        {
            fprintf(stderr, "\nError: Invalid source IP address: %s\n", iphdr_o.saddr);
            exit(EXIT_FAILURE);
        }

        if((iphdr_o.v4daddr_nbo = libnet_name2addr4(packet_d, iphdr_o.daddr, 1)) == -1)
        {
            fprintf(stderr, "\nError: Invalid source IP address: %s\n", iphdr_o.daddr);
            exit(EXIT_FAILURE);
        }
    }

    if(!iphdr_o.id || rand_id)
    {
        rand_id = 1; 
        iphdr_o.id = libnet_get_prand(LIBNET_PRu16);
    }

    for(packet_c = 1; packet_c < count + 1; packet_c++)
    {
        if(iphdr_o.version == 4)
            if(iphdr_o.rand_saddr)
                iphdr_o.saddr = packit_retrieve_rand_v4_addr();

        if(iphdr_o.protocol == IPPROTO_TCP)
        {
            strcpy(proto, "TCP");

            packet_d = packit_build_tcp(packet_d);
            ip_proto = IPPROTO_TCP;
            hdrlen = PACKIT_TCP_H;

            if(packet_c == 1)
            {
                fprintf(stderr, "IPv%d %s %s.%d > %s.%d: ",
                    iphdr_o.version,
                    proto,
                    iphdr_o.saddr,
                    thdr_o.th_sport,
                    iphdr_o.daddr, 
                    thdr_o.th_dport);


                if(iphdr_o.protocol == IPPROTO_TCP)
                {
                    if(thdr_o.th_syn)
                        strncat(flags, "S", 1);

                    if(thdr_o.th_fin)
                        strncat(flags, "F", 1);

                    if(thdr_o.th_rst)
                        strncat(flags, "R", 1);

                    if(thdr_o.th_psh)
                        strncat(flags, "P", 1);

                    if(strlen(flags) < 1)
                        fprintf(stderr, ". ");
                    else
                        fprintf(stderr, "%s ", flags);

                    if(thdr_o.th_seqn)
                        fprintf(stderr, "seq %ld ", thdr_o.th_seqn);

                    if(thdr_o.th_ack || thdr_o.th_ackn)
                        fprintf(stderr, "ack ");

                    if(thdr_o.th_ackn)
                        fprintf(stderr, "%ld ", thdr_o.th_ackn);

                    fprintf(stderr, "win %d ", thdr_o.th_win);

                    if(thdr_o.th_urg)
                        fprintf(stderr, "urg ");

                    if(thdr_o.th_urp)
                        fprintf(stderr, "%d ", thdr_o.th_urp);
                }

                if(init_type)
                    fprintf(stderr, "raw4 ");
                else
                    fprintf(stderr, "link ");

                fprintf(stderr, "\n");
                fprintf(stderr, "Writing packet(s) (%d): ", count);
            }
        }
        else if(iphdr_o.protocol == IPPROTO_UDP)
        {
            strcpy(proto, "UDP");

            packet_d = packit_build_udp(packet_d);
            ip_proto = IPPROTO_UDP;
            hdrlen = PACKIT_UDP_H;

            if(packet_c == 1)
            {
                fprintf(stderr, "ipv%d %s %s.%d > %s.%d ",
                    iphdr_o.version,
                    proto,
                    iphdr_o.saddr,
                    uhdr_o.uh_sport,
                    iphdr_o.daddr,
                    uhdr_o.uh_dport);

                fprintf(stderr, "\n");
                fprintf(stderr, "writing packet(s) (%d): ", count);
            }
        }
        else if(iphdr_o.protocol == IPPROTO_ICMP)
        {
            strcpy(proto, "ICMP");

            packet_d = packit_build_icmp(packet_d);
            ip_proto = IPPROTO_ICMP;

            switch(ihdr_o.ih_type)
            {
                case 0: case 8:
                    hdrlen = PACKIT_ICMPV4_ECHO_H;
                    break;

                case 3: 
                    hdrlen = PACKIT_ICMPV4_UNREACH_H;
                    break;

                case 5:
                    hdrlen = PACKIT_ICMPV4_REDIRECT_H;
                    break;

                case 11:
                    hdrlen = PACKIT_ICMPV4_TIMXCEED_H;
                    break;

                case 13: case 14:
                    hdrlen = PACKIT_ICMPV4_TS_H;
                    break;

                case 17: case 18:
                    hdrlen = PACKIT_ICMPV4_MASK_H;
                    break;
            }

            if(packet_c == 1)
            {
                fprintf(stderr, "ipv%d %s %s > %s: ",
                    iphdr_o.version,
                    proto,
                    iphdr_o.saddr,
                    iphdr_o.daddr);

                fprintf(stderr, "\n");
                fprintf(stderr, "writing packet(s) (%d): ", count);
            }
        }

        if(libnet_build_ipv4(
            PACKIT_IPV4_H + hdrlen + payload_s,
            iphdr_o.tos,
            iphdr_o.id,
            0,
            iphdr_o.ttl,
            ip_proto,
            0,
            iphdr_o.v4saddr_nbo,
            iphdr_o.v4daddr_nbo,
            NULL,
            0,
            packet_d,
            0) == -1)
        {
            fprintf(stderr, "\nError: Unable to build IPv4 header: %s\n", libnet_geterror(packet_d));
            exit(EXIT_FAILURE);
        }

        if(init_type == 0)
        {
           if(libnet_build_ethernet(
                ehdr_o.eh_daddr,
                ehdr_o.eh_saddr,
                ETHERTYPE_IP,
                NULL,
                0,
                packet_d,
                0) == -1)
            {
                fprintf(stderr, "\nError: Unable to build Ethernet header: %s\n", libnet_geterror(packet_d));
                exit(EXIT_FAILURE);
            }
        }

        if(libnet_write(packet_d) == -1)
        {
            fprintf(stderr, "\nError: Unable to write packet to network: %s\n", libnet_geterror(packet_d));
            exit(EXIT_FAILURE); 
        }

        fprintf(stderr, ".");

        /* we shouldn't have to use this */
        libnet_clear_packet(packet_d);

        if(count > 1 && burst && (packet_c % burst) == 0)
            sleep(1);
        else if(count > 1 && count != packet_c && !burst)
            sleep(interval);
    }
  
    libnet_stats(packet_d, &ln_stats);

    fprintf(stderr, "\n\n");
    fprintf(stderr, "------ Injection Statistics ------\n");
    fprintf(stderr,
        "Packet(s) sent: %ld  "
        "Packet errors: %ld  "
        "Bytes written: %ld  ",
        ln_stats.packets_sent, ln_stats.packet_errors, ln_stats.bytes_written);

    libnet_destroy(packet_d);

    fprintf(stderr, "\n");

    return;
}
